/*****************************************************************************/
/*
                                 DECnet.c

This module implements a full multi-threaded, AST-driven, script execution via
DECnet.  Primarily this is to provide a substantially OSU-compatible scripting
enviroment, but does support WASD-style CGI scripting tranparently over DECnet.
This code is functional but does not pretend to be highly optimized.

Two modes are supported.


WASD CGI MODE
-------------
These scripts behave in exactly the same manner as subprocess-based, standard
WASD CGI scripts. A stream of DCL commands create the CGI variable environment
and execute the script. The output stream can be CGI-compliant or return a
full HTTP response stream.

This is the default mode. If a mapped script contains a DECnet node component
but without a "TASK=" string the CGI mode is used.  It is also the mode for
any "TASK=" string containing "CGIWASD".


OSU MODE
--------
This mode emulates the OSU DECnet scripting environment.  It is based in large
part on reverse engineering the OSU 'script_execute.c' module, and lots of
trial-and-error!

This is the mode used when the mapped script contains a DECnet component and a
"TASK=" string that does not contain "WASD" as the final four characters.  In
general the string will be "TASK=WWWEXEC", the basic OSU script executive.
Any other task name will also be assumed to be OSU compliant.


MAPPING
-------
DECnet scripting is enabled for individual scripts by including a DECnet
component in the result of a "script" or "exec" mapping rule.

Examples ...

  exec /FRODO/* /FRODO::/cgi-bin/*
  exec /frodo/* /frodo::"task=cgiwasd"/cgi-bin/*

executes any WASD CGI script specified on node FRODO
(both forms result in the same script execution)

  exec /bilbo/osu/* /BILBO::"TASK=WWWEXEC"/*

executes any OSU script specified on node BILBO


VERSION HISTORY
---------------
16-DEC-97  MGD  initial development for v5.0
*/
/*****************************************************************************/

/* standard C header files */
#include <stdio.h>
#include <ctype.h>
#include <string.h>

/* VMS related header files */
#include <descrip.h>
#include <iodef.h>
#include <libdef.h>
#include <ssdef.h>
#include <stsdef.h>

/* application header files */
#include "wasd.h"
#include "auth.h"
#include "CGI.h"
#include "DECnet.h"
#include "error.h"
#include "httpd.h"
#include "mapurl.h"
#include "msg.h"
#include "net.h"
#include "support.h"

/**********/
/* macros */
/**********/

#define DECNET_CONTENT_MAX 1024

/* space for adding file types (e.g. ".COM") onto script specifications */
#define FIND_SCRIPT_OVERHEAD 48

#define DECNET_TASK_CGI "CGIWASD"
#define DECNET_TASK_OSU "WWWEXEC"

#define SCRIPT_CGI      1
#define SCRIPT_OSU      2

#define CGI_BEGIN       1
#define CGI_DCL         2
#define CGI_OUTPUT      3

#define OSU_BEGIN       1
#define OSU_DIALOG      2
#define OSU_DNET_HDR    3
#define OSU_DNET_INPUT  4
#define OSU_DNET_XLATE  5
#define OSU_OUTPUT_RAW  6
#define OSU_OUTPUT_REC  7
#define OSU_DNETTEXT    8

/******************/
/* global storage */
/******************/

char  ErrorOsuImplementation [] = "Possible OSU implementation problem!",
      ErrorOsuNoInvCache [] =
         "OSU &quot;invalidate cache&quot; dialog not implemented.",
      ErrorOsuNoManage [] = "OSU &quot;manage&quot; dialog not implemented.";

/********************/
/* external storage */
/********************/

#ifdef DBUG
extern boolean Debug;
#else
#define Debug 0 
#endif

extern int  NetReadBufferSize;
extern int  OutputBufferSize;
extern char  DclCgiVariablePrefix[];
extern char  ErrorSanityCheck[];
extern char  HttpProtocol[];
extern char  SoftwareID[];
extern struct AccountingStruct  Accounting;
extern struct ConfigStruct  Config;
extern struct MsgStruct  Msgs;

/*****************************************************************************/
/*
Returns true if there is a DECnet specific problem, else returns false.
*/ 
 
#ifdef __DECC
#pragma inline(DECnetNotOk)
#endif

boolean DECnetNotOk (int StatusValue)
{
   if (VMSok (StatusValue)) return (false);
   if (StatusValue == SS$_ABORT ||
       StatusValue == SS$_CANCEL ||
       StatusValue == SS$_LINKABORT ||
       StatusValue == SS$_LINKDISCON ||
       StatusValue == SS$_LINKEXIT)
      return (true);
   else
      return (false);
}

/*****************************************************************************/
/*
Initiate a connection to a DECnet node using the connection details specified
in 'ConnectString'.
*/ 
 
int DECnetBegin
(
struct RequestStruct *rqptr,
void *NextTaskFunction,
char *MappedScript
)
{
   static $DESCRIPTOR (NetDeviceDsc, "_NET:");
   static $DESCRIPTOR (NcbDsc, "");

   register char  *cptr, *sptr, *zptr;
   register struct DECnetTaskStruct  *tkptr;

   int  status;
   void  *AstFunctionPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DECnetBegin() |%s|\n", MappedScript);

   if (!rqptr->AccountingDone)
      rqptr->AccountingDone = ++Accounting.DoDECnetCount;

   /* set up the task structure */
   if (rqptr->DECnetTaskPtr == NULL)
   {
      rqptr->DECnetTaskPtr = tkptr = (struct DECnetTaskStruct*)
         VmGetHeap (rqptr, sizeof(struct DECnetTaskStruct));
   }
   else
   {
      tkptr = rqptr->DECnetTaskPtr;
      memset (tkptr, 0, sizeof(struct DECnetTaskStruct));
   }
   tkptr->NextTaskFunction = NextTaskFunction;

   cptr = MappedScript;
   zptr = (sptr = tkptr->MappedScript) + sizeof(tkptr->MappedScript);
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if (sptr >= zptr)
   {
      ErrorGeneralOverflow (rqptr, FI_LI);
      DECnetEnd (rqptr);
      return;
   }
   *sptr = '\0';

   /*******************************/
   /* generate the connect string */
   /*******************************/

   zptr = (sptr = tkptr->ConnectString) + sizeof(tkptr->ConnectString);
   for (cptr = tkptr->MappedScript;
        *cptr && *cptr != ':' && sptr < zptr;
        *sptr++ = *cptr++);
   if (cptr[0] == ':' && cptr[1] == ':')
   {
      /* isolate the DECnet component from the script file specification */
      if (sptr < zptr) *sptr++ = *cptr++;
      if (sptr < zptr) *sptr++ = *cptr++;
      if (*cptr == '\"')
      {
         /* mapping rule is supplying the DECnet task information */
         if (sptr < zptr) *sptr++ = *cptr++;
         while (*cptr && *cptr != '\"' && sptr < zptr) *sptr++ = *cptr++;
         if (sptr < zptr) *sptr++ = *cptr++;
         *sptr = '\0';
         tkptr->ScriptPtr = cptr + 1;
         DECnetSetDialog (tkptr);
      }
      else
      {
         /* supply the default WASD CGI DECnet task */
         tkptr->ScriptPtr = cptr;
         cptr = "\"TASK=";
         while (*cptr && sptr < zptr) *sptr++ = *cptr++;
         cptr = DECNET_TASK_CGI;
         while (*cptr && sptr < zptr) *sptr++ = *cptr++;
         if (sptr < zptr) *sptr++ = '\"';
         if (sptr >= zptr)
         {
            ErrorGeneralOverflow (rqptr, FI_LI);
            DECnetEnd (rqptr);
            return;
         }
         *sptr = '\0';
         tkptr->ScriptType = SCRIPT_CGI;
      }
      if (Debug)
         fprintf (stdout, "|%s|%s|\n", tkptr->ConnectString, tkptr->ScriptPtr);

      /* apply that to the network Connect Block descriptor */
      NcbDsc.dsc$w_length = sptr - tkptr->ConnectString;
      NcbDsc.dsc$a_pointer = tkptr->ConnectString;
   }
   else
   {
      /* shouldn't be in this function if no DECnet in specification! */
      ErrorGeneral (rqptr, ErrorSanityCheck, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   /**************************/
   /* connect to DECnet task */
   /**************************/

   status = sys$assign (&NetDeviceDsc, &tkptr->DECnetChannel, 0, 0);
   if (Debug) fprintf (stdout, "lib$assign() %%X%08.08X\n", status);
   if (VMSnok (status))
   {
      rqptr->ErrorTextPtr = rqptr->ScriptName;
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   if (Debug)
      fprintf (stdout, "|%*.*s|\n",
               NcbDsc.dsc$w_length, NcbDsc.dsc$w_length, NcbDsc.dsc$a_pointer);

   status = sys$qio (0, tkptr->DECnetChannel, IO$_ACCESS,
                     &tkptr->DECnetConnectIOsb, &DECnetConnectAst, rqptr,
                     0, &NcbDsc, 0, 0, 0, 0);
   if (Debug) fprintf (stdout, "sys$qio() %%X%08.08X\n", status);
   if (VMSok (status)) return;

   tkptr->DECnetConnectIOsb.Status = status;
   SysDclAst (&DECnetConnectAst, rqptr);
}

/*****************************************************************************/
/*
Examine the connect string to determine the scripting dialog required.  Set a
flag in the task structure to indicate the type of dialog.
*/ 
 
int DECnetSetDialog (struct DECnetTaskStruct *tkptr)

{
   register char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "DECnetSetDialog() |%s|\n", tkptr->ConnectString);

   cptr = tkptr->ConnectString;
   while (*cptr && *cptr != '\"')
   {
      if (!(strsame (cptr, "\"TASK=", 6) || strsame (cptr, "\"0=", 3)))
      {
         cptr++;
         continue;
      }
      if (cptr[1] == '0')
         cptr += 3;
      else
         cptr += 6;

      if (strsame (cptr, DECNET_TASK_CGI, sizeof(DECNET_TASK_CGI)-1) &&
          cptr[sizeof(DECNET_TASK_CGI)] == '\"')
      {
         tkptr->ScriptType = SCRIPT_CGI;
         if (Debug) fprintf (stdout, "dialog: %d\n", tkptr->ScriptType);
         return;
      }
   }

   /* any task name not otherwise recognized becomes an OSU-based script */
   tkptr->ScriptType = SCRIPT_OSU;
   if (Debug) fprintf (stdout, "dialog: %d\n", tkptr->ScriptType);
}

/*****************************************************************************/
/*
*/ 
 
int DECnetEnd (struct RequestStruct *rqptr)

{
   register struct DECnetTaskStruct  *tkptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "DECnetEnd() %d %d\n",
               rqptr, rqptr->DECnetTaskPtr->QueuedDECnetIO);

   tkptr = rqptr->DECnetTaskPtr;

   if (tkptr->QueuedDECnetIO)
   {
      sys$cancel (tkptr->DECnetChannel);
      return;
   }

   status = sys$dassgn (tkptr->DECnetChannel);
   if (Debug) fprintf (stdout, "sys$dassgn() %%X%08.08X\n", status);

   /* declare the next task */
   SysDclAst (tkptr->NextTaskFunction, rqptr);
}

/*****************************************************************************/
/*
Th connection request to the specified DECnet task has completed.  Check the
status and report any error.  If no error then initialize the required script
dialog and call the appropriate function to begin it.
*/ 
 
int DECnetConnectAst (struct RequestStruct *rqptr)

{
   register struct DECnetTaskStruct  *tkptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "DECnetConnectAst() %d %%X%08.08X\n",
               rqptr, rqptr->DECnetTaskPtr->DECnetConnectIOsb);

   tkptr = rqptr->DECnetTaskPtr;

   if (VMSnok (tkptr->DECnetConnectIOsb.Status))
   {
      rqptr->ErrorTextPtr = rqptr->ScriptName;
      ErrorVmsStatus (rqptr, tkptr->DECnetConnectIOsb.Status, FI_LI);
      DECnetEnd (rqptr);
      return;
   }

   switch (tkptr->ScriptType)
   {
      case SCRIPT_CGI :
         Accounting.DoDECnetCgiCount++;
         rqptr->CgiScript = true;
         tkptr->CgiDialogState = CGI_BEGIN;
         DECnetCgiDialog (rqptr);
         return;

      case SCRIPT_OSU :
         Accounting.DoDECnetOsuCount++;
         tkptr->OsuDialogState = OSU_BEGIN;
         DECnetOsuDialog (rqptr);
         return;

      default :
         ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);
   }
}

/*****************************************************************************/
/*
Write a record to the remote DECnet task.  AST function will process any
errors at all.
*/ 
 
int DECnetWrite
(
struct RequestStruct *rqptr,
char *DataPtr,
int DataLength
)
{
   register struct DECnetTaskStruct  *tkptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
   {
      int  len;
      if (DataLength >= 0) len = DataLength; else len = strlen(DataPtr);
      fprintf (stdout, "DECnetWrite()\n");
      fprintf (stdout, "%d |%*.*s|\n", DataLength, len, len, DataPtr);
   }

   tkptr = rqptr->DECnetTaskPtr;

   if (DataLength < 0) DataLength = strlen (DataPtr);

   tkptr->QueuedDECnetIO++;
   status = sys$qio (0, tkptr->DECnetChannel, IO$_WRITEVBLK,
                     &tkptr->DECnetWriteIOsb, &DECnetWriteAst, rqptr,
                     DataPtr, DataLength, 0, 0, 0, 0);
   if (Debug) fprintf (stdout, "sys$qio() %%X%08.08X\n", status);
   if (VMSok (status)) return;

   tkptr->DECnetWriteIOsb.Status = status;
   SysDclAst (&DECnetWriteAst, rqptr);
}

/*****************************************************************************/
/*
*/ 
 
int DECnetWriteAst (struct RequestStruct *rqptr)

{
   register struct DECnetTaskStruct  *tkptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "DECnetWriteAst() %d %d %%X%08.08X\n",
               rqptr, rqptr->DECnetTaskPtr->QueuedDECnetIO,
               rqptr->DECnetTaskPtr->DECnetWriteIOsb.Status);

   tkptr = rqptr->DECnetTaskPtr;

   if (tkptr->QueuedDECnetIO) tkptr->QueuedDECnetIO--;

   if (DECnetNotOk (tkptr->DECnetWriteIOsb.Status))
   {
      if (!tkptr->QueuedDECnetIO ||
          tkptr->DECnetReadIOsb.Status != SS$_CANCEL)
         DECnetEnd (rqptr);
      return;
   }

   /* reusing the IOsb, expect it to occasionally have been initialized! */
   if (!tkptr->DECnetWriteIOsb.Status)
      tkptr->DECnetWriteIOsb.Status = SS$_NORMAL;
   if (VMSnok (tkptr->DECnetWriteIOsb.Status))
   {
      rqptr->ErrorTextPtr = rqptr->ScriptName;
      ErrorVmsStatus (rqptr, tkptr->DECnetWriteIOsb.Status, FI_LI);
      DECnetEnd (rqptr);
      return;
   }
}

/*****************************************************************************/
/*
Write a chunk of the request header or body to the script. When the content
data is exhausted write an empty record as and end-of-data marker. This must
be checked for and detected by the script as there is no end-of-file
associated with this stream!
*/ 
 
int DECnetWriteRequest (struct RequestStruct *rqptr)

{
   register char  *cptr;
   register struct DECnetTaskStruct  *tkptr;

   int  status,
        DataLength;
   char  *DataPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "DECnetWriteRequest() %d %d\n",
               rqptr->DECnetTaskPtr->ContentPtr,
               rqptr->DECnetTaskPtr->ContentCount);

   tkptr = rqptr->DECnetTaskPtr;

   if (tkptr->ContentPtr == NULL)
   {
      /*********************************/
      /* initalize the data to be sent */
      /*********************************/

      switch (tkptr->ScriptType)
      {
         case SCRIPT_CGI :

            /* request body */
            if (rqptr->ContentBufferPtr == NULL)
            {
               tkptr->ContentPtr = "";
               tkptr->ContentCount = 0;
            }
            else
            {
               tkptr->ContentPtr = rqptr->ContentBufferPtr;
               tkptr->ContentCount = rqptr->ContentLength;
            }
            break;

         case SCRIPT_OSU :

            switch (tkptr->OsuDialogState)
            {
               case OSU_DNET_HDR :

                  /* request header (excluding first line) */
                  cptr = rqptr->RequestHeaderPtr;
                  while (NOTEOL(*cptr)) cptr++;
                  if (*cptr == '\r') cptr++;
                  if (*cptr == '\n') cptr++;
                  tkptr->ContentPtr = cptr;
                  tkptr->ContentCount = -1;
                  break;

               case OSU_DNET_INPUT :

                  /* request body */
                  if (rqptr->ContentBufferPtr == NULL)
                  {
                     tkptr->ContentPtr = "";
                     tkptr->ContentCount = 0;
                  }
                  else
                  {
                     tkptr->ContentPtr = rqptr->ContentBufferPtr;
                     tkptr->ContentCount = rqptr->ContentLength;
                  }
                  break;

               default :
                  ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);
            }
            break;

         default :
            ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);

      }
      tkptr->ContentCountZero = 0;
   }

   /*************/
   /* send data */
   /*************/

   if (tkptr->ScriptType == SCRIPT_OSU &&
       tkptr->OsuDialogState == OSU_DNET_HDR)
   {
      /* OSU gets the header line-by-line (excluding carriage-control) */
      for (cptr = DataPtr = tkptr->ContentPtr; NOTEOL(*cptr); cptr++);
      DataLength = cptr - DataPtr;
      if (*cptr == '\r') cptr++;
      if (*cptr == '\n') cptr++;
      /* if end-of-header blank line, indicate this */
      if (*DataPtr == '\r' || *DataPtr == '\n')
         tkptr->ContentPtr = NULL;
      else
         tkptr->ContentPtr = cptr;
   }
   else
   if (!tkptr->ContentCount)
   {
      /* write end-of-output empty record (blank line) */
      DataPtr = "";
      DataLength = 0;
      /* indicate we're all done */
      tkptr->ContentPtr = NULL;
   }
   else
   {
      /* request body is sent in "chunks" */
      DataPtr = tkptr->ContentPtr;
      if (tkptr->ContentCount > DECNET_CONTENT_MAX)
         DataLength = DECNET_CONTENT_MAX;
      else
         DataLength = tkptr->ContentCount;
      tkptr->ContentCount -= DataLength;
      tkptr->ContentPtr += DataLength;
   }
   if (Debug) fprintf (stdout, "%d %d\n", DataPtr, DataLength);

   status = sys$qio (0, tkptr->DECnetChannel, IO$_WRITEVBLK,
                     &tkptr->DECnetWriteIOsb, &DECnetWriteRequestAst, rqptr,
                     DataPtr, DataLength, 0, 0, 0, 0);
   if (Debug) fprintf (stdout, "sys$qio() %%X%08.08X\n", status);
   if (VMSok (status)) return;

   /* report the sys$qio() problem via the AST routine */
   tkptr->DECnetWriteIOsb.Status = status;
   SysDclAst (&DECnetWriteRequestAst, rqptr);
}

/*****************************************************************************/
/*
A content write has completed.  Check the IO status block for errors.  If no
error then check for more content to be written, if so the write it!
*/ 
 
int DECnetWriteRequestAst (struct RequestStruct *rqptr)

{
   register struct DECnetTaskStruct  *tkptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "DECnetWriteRequestAst() %d %d %%X%08.08X %d\n",
               rqptr, rqptr->DECnetTaskPtr->QueuedDECnetIO,
               rqptr->DECnetTaskPtr->DECnetWriteIOsb.Status,
               rqptr->DECnetTaskPtr->DECnetWriteIOsb.Count);

   tkptr = rqptr->DECnetTaskPtr;

   if (tkptr->QueuedDECnetIO) tkptr->QueuedDECnetIO--;

   if (DECnetNotOk (tkptr->DECnetWriteIOsb.Status))
   {
      if (!tkptr->QueuedDECnetIO ||
          tkptr->DECnetReadIOsb.Status != SS$_CANCEL)
         DECnetEnd (rqptr);
      return;
   }

   /* reusing the one IOsb, expect it occasionally to be initialized! */
   if (!tkptr->DECnetWriteIOsb.Status)
      tkptr->DECnetWriteIOsb.Status = SS$_NORMAL;
   if (VMSnok (tkptr->DECnetWriteIOsb.Status))
   {
      rqptr->ErrorTextPtr = rqptr->ScriptName;
      ErrorVmsStatus (rqptr, tkptr->DECnetWriteIOsb.Status, FI_LI);
      DECnetEnd (rqptr);
      return;
   }

   if (tkptr->ContentPtr == NULL)
   {
      /* end-of-output empty record has been written ... finished! */
      if (tkptr->ScriptType == SCRIPT_OSU)
      {
         /* back from <DNETHDR> or <DNETINPUT> into dialog */
         tkptr->OsuDialogState = OSU_DIALOG;
         DECnetRead (rqptr);
      }
   }
   else
   {
      /* write more data */
      DECnetWriteRequest (rqptr);
   }
}

/*****************************************************************************/
/*
Queue a read from the remote DECnet task. AST function will process any errors
at all.
*/ 
 
int DECnetRead (struct RequestStruct *rqptr)

{
   register struct DECnetTaskStruct  *tkptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DECnetRead()\n");

   tkptr = rqptr->DECnetTaskPtr;

   if (rqptr->OutputBufferPtr == NULL)
   {
      /* initialize output buffer */
      NetWriteBuffered (rqptr, 0, NULL, 0);
   }

   tkptr->QueuedDECnetIO++;
   status = sys$qio (0, tkptr->DECnetChannel, IO$_READVBLK | IO$M_MULTIPLE,
                     &tkptr->DECnetReadIOsb, &DECnetReadAst, rqptr,
                     rqptr->OutputBufferPtr, OutputBufferSize,
                     0, 0, 0, 0);
   if (Debug) fprintf (stdout, "sys$qio() %%X%08.08X\n", status);
   if (VMSok (status)) return;

   tkptr->DECnetReadIOsb.Status = status;
   SysDclAst (&DECnetReadAst, rqptr);
}

/*****************************************************************************/
/*
The queued read from the remote DECnet task has completed.  Check for any
errors reported in the IO status block.  If no errors call the function
appropriate for processing the dialog being used by this script.
*/ 
 
int DECnetReadAst (struct RequestStruct *rqptr)

{
   register struct DECnetTaskStruct  *tkptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "DECnetReadAst() %d %d %%X%08.08X %d bytes\n",
               rqptr, rqptr->DECnetTaskPtr->QueuedDECnetIO,
               rqptr->DECnetTaskPtr->DECnetReadIOsb.Status,
               rqptr->DECnetTaskPtr->DECnetReadIOsb.Count);

   tkptr = rqptr->DECnetTaskPtr;

   if (tkptr->QueuedDECnetIO) tkptr->QueuedDECnetIO--;

   if (DECnetNotOk (tkptr->DECnetReadIOsb.Status))
   {
      if (!tkptr->QueuedDECnetIO ||
          tkptr->DECnetReadIOsb.Status != SS$_CANCEL)
         DECnetEnd (rqptr);
      return;
   }

   if (VMSnok (tkptr->DECnetReadIOsb.Status))
   {
      rqptr->ErrorTextPtr = rqptr->ScriptName;
      ErrorVmsStatus (rqptr, tkptr->DECnetReadIOsb.Status, FI_LI);
      DECnetEnd (rqptr);
      return;
   }

   rqptr->OutputBufferPtr[tkptr->DECnetReadIOsb.Count] = '\0';

   switch (tkptr->ScriptType)
   {
      case SCRIPT_CGI :
         DECnetCgiDialog (rqptr);
         return;

      case SCRIPT_OSU :
         DECnetOsuDialog (rqptr);
         return;

      default :
         ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);
   }
}

/*****************************************************************************/
/*
Dialog for a WASD standard CGI DECnet-based script. These scripts behave
identically to subprocess-based standard CGI scripts. The dialog comprises two
phases. The first sends a stream of records to the CGI DECnet task. These
records comprise DCL commands, used to set up the CGI environment and to
execute the script. The second phase uses the generic CGI modules functions to
process CGI script output, writing it to the client.
*/ 
 
DECnetCgiDialog (struct RequestStruct *rqptr)

{
   static char  GotoDoIt [] = "GOTO DOIT";
   static $DESCRIPTOR (DefineEofFaoDsc, "DEFINE/PROCESS CGIEOF \"!AZ\"");
   static $DESCRIPTOR (DclCommandDsc, "");

   register char  *cptr;
   register struct DECnetTaskStruct  *tkptr;

   int  status,
        value;
   unsigned short  Length;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DECnetCgiDialog()\n");

   tkptr = rqptr->DECnetTaskPtr;

   if (tkptr->CgiDialogState == CGI_BEGIN)
   {
      /***********************/
      /* look for the script */
      /***********************/

      status = DECnetFindScript (rqptr);
      if (Debug) fprintf (stdout, "find script %%X%08.08X\n", status);

      /* retry indicates the search is still underway */
      if (status == SS$_RETRY) return;

      if (VMSnok (status))
      {
         /* fudge the status so the write AST reports the problem */
         tkptr->DECnetWriteIOsb.Status = status;
         SysDclAst (&DECnetWriteAst, rqptr);
         return;
      }

      /* must have found the script, set up the DCL/CGI environment */
      tkptr->CgiDialogState = CGI_DCL;
   }

   if (tkptr->CgiDialogState == CGI_DCL)
   {
      /*****************/
      /* define CGIEOF */
      /*****************/

      /* always generate a new EOF string for standard CGI scripts */
      CgiEof (tkptr->DECnetEof, &tkptr->DECnetEofLength);
      rqptr->CgiEofPtr = tkptr->DECnetEof;
      rqptr->CgiEofLength = tkptr->DECnetEofLength;

      DclCommandDsc.dsc$a_pointer = tkptr->DclCommand;
      DclCommandDsc.dsc$w_length = sizeof(tkptr->DclCommand)-1;
      sys$fao (&DefineEofFaoDsc, &Length, &DclCommandDsc, tkptr->DECnetEof);
      tkptr->DclCommand[Length] = '\0';
      DECnetWrite (rqptr, tkptr->DclCommand, Length);

      /*****************/
      /* CGI variables */
      /*****************/

      if (VMSnok (status = CgiGenerateVariables (rqptr, CGI_VARIABLE_DCL)))
      {
         /* fudge the status so the write AST reports the problem */
         tkptr->DECnetWriteIOsb.Status = status;
         SysDclAst (&DECnetWriteAst, rqptr);
         return;
      }

      cptr = rqptr->CgiBufferPtr;
      for (;;)
      {
         if (!(Length = *(short*)cptr)) break;
         DECnetWrite (rqptr, cptr+sizeof(short), Length-1);
         cptr += Length + sizeof(short);
      }

      /*******************************/
      /* DCL command/procedure/image */
      /*******************************/

      if (tkptr->CgiScriptForeignVerb[0])
         DECnetWrite (rqptr, tkptr->CgiScriptForeignVerb, -1);

      DECnetWrite (rqptr, GotoDoIt, sizeof(GotoDoIt)-1);

      DECnetWrite (rqptr, tkptr->CgiScriptDcl, -1);

      if (rqptr->HttpMethod == HTTP_METHOD_POST ||
          rqptr->HttpMethod == HTTP_METHOD_PUT)
      {
         /* begin to write the request content (body) */
         DECnetWriteRequest (rqptr);
      }

      /*******************************/
      /* begin to read script output */
      /*******************************/

      /* now into dialog state */
      tkptr->CgiDialogState = CGI_OUTPUT;

      DECnetRead (rqptr);
      return;
   }
   else
   /* CGI_OUTPUT */
   {
      /*************************/
      /* process script output */
      /*************************/

      value = CgiOutput (rqptr,
                         rqptr->OutputBufferPtr,
                         tkptr->DECnetReadIOsb.Count);

      if (value == CGI_OUTPUT_TERMINATE)
      {
         /* terminate processing */
         DECnetEnd (rqptr);
         return;
      }

      if (value == CGI_OUTPUT_ABSORB)
      {
         /* absorb output, queue the next read from the script */
         DECnetRead (rqptr);
         return;
      }

      /* otherwise it's the count of bytes in the output buffer */
      NetWrite (rqptr, &DECnetRead, rqptr->OutputBufferPtr, value);
      return;
   }
}

/*****************************************************************************/
/*
Use the DECnet CGI task to search for the script file. A DCL command is
constructed with the F$SEARCH() lexical and written to the DECnet task which
then executes it, writing the result as an output record which is read by
DECnetRead(), passed to DECnetCgiDialog() and passed back to this same
function for result assessment. If not found and if no explicit file type,
step through each of ".COM", ".EXE" and then any configuration specified file
types and run-time environments, repeating this until found or until all file
types exhausted, at which time a "script not found" report is generated. If
found the DCL required to execute the script is generated in
'tkptr->CgiScriptForeignVerb' and 'tkptr->CgiScriptDcl'.
*/ 

int DECnetFindScript (struct RequestStruct *rqptr)

{
   static $DESCRIPTOR (SearchFaoDsc, "WRITE NET$LINK F$SEARCH(\"!AZ\")");
   static $DESCRIPTOR (DclCommandDsc, "");

   register int  idx;
   register char  *cptr, *sptr, *zptr;
   register struct DECnetTaskStruct  *tkptr;

   int  status;
   unsigned short  Length;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DECnetFindScript()\n");

   tkptr = rqptr->DECnetTaskPtr;

   if (tkptr->FindScriptCount && tkptr->DECnetReadIOsb.Count)
   {
      rqptr->OutputBufferPtr[tkptr->DECnetReadIOsb.Count] = '\0';
      if (Debug)
      {
         fprintf (stdout, "f$search() %d |%s| %d |%s|\n",
                  tkptr->DECnetReadIOsb.Count, rqptr->OutputBufferPtr,
                  tkptr->FindScriptTotalLength, tkptr->FindScript);
      }

      if (strsame (rqptr->OutputBufferPtr,
                   tkptr->FindScript,
                   tkptr->FindScriptTotalLength))
      {
         /*************/
         /* found it! */
         /*************/

         if (Debug) fprintf (stdout, "FOUND!\n");
         tkptr->CgiScriptDcl[0] = tkptr->CgiScriptForeignVerb[0] = '\0';

         cptr = tkptr->RunTimePtr;
         if (*(unsigned short*)cptr == '@\0')
         {
            /* DCL procedure */
            sptr = tkptr->CgiScriptDcl;
            *sptr++ = '@';
         }
         else
         if (*(unsigned short*)cptr == '$\0')
         {
            /* execute an image */
            sptr = tkptr->CgiScriptDcl;
            for (cptr = "RUN "; *cptr; *sptr++ = *cptr++);
         }
         else
         {
            if (*cptr == '@' || *cptr == '$')
            {
               /* foreign-verb DCL procedure or executable, create the verb */
               sptr = tkptr->CgiScriptForeignVerb;
               cptr++;
               while (*cptr) *sptr++ = *cptr++;
               *sptr = '\0';

               /* now place it as the verb before the script file */
               sptr = tkptr->CgiScriptDcl;
               while (*cptr && *cptr != '=') *sptr++ = *cptr++;
               *sptr++ = ' ';
               *sptr = '\0';
            }
            else
            {
               /* verb already exists, place before the script file */
               sptr = tkptr->CgiScriptDcl;
               while (*cptr) *sptr++ = *cptr++;
               *sptr++ = ' ';
               *sptr = '\0';
            }
         }

         cptr = tkptr->FindScript;
         while (*cptr) *sptr++ = *cptr++;
         *sptr = '\0';

         if (Debug)
            fprintf (stdout, "|%s|%s|\n",
                     tkptr->CgiScriptForeignVerb, tkptr->CgiScriptDcl);
         return (SS$_NORMAL);
      }

      /* not our dialog, force not found report */
      tkptr->FindScriptCount = 999999;
   }

   /*******************/
   /* not found (yet) */
   /*******************/

   if (tkptr->FindScriptCount == 1) 
   {
      /* explicitly supplied file type was not found */
      rqptr->ResponseStatusCode = 404;
      ErrorGeneral (rqptr, MsgFor(rqptr,MSG_SCRIPT_NOT_FOUND), FI_LI);
      return (STS$K_ERROR);
   }

   /* if started searching then search for the next one */
   if (tkptr->FindScriptCount > 1) tkptr->FindScriptCount++;

   if (!tkptr->FindScriptCount)
   {
      /****************/
      /* initial call */
      /****************/

      /* get the script file name (minus all DECnet-related stuff) */
      zptr = (sptr = tkptr->FindScript) +
             sizeof(tkptr->FindScript) - FIND_SCRIPT_OVERHEAD;
      for (cptr = tkptr->MappedScript; *cptr && *cptr != ':'; cptr++);
      if (*cptr == ':') cptr++;
      if (*cptr == ':') cptr++;
      if (*cptr == '\"')
      {
         /* task specification string */
         cptr++;
         while (*cptr != '\"') cptr++;
         if (*cptr == '\"') cptr++;
      }
      /* here we go with the file name */
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      if (sptr >= zptr)
      {
         ErrorGeneralOverflow (rqptr, FI_LI);
         return (STS$K_ERROR);
      }
      *sptr = '\0';
      tkptr->FindScriptLength = sptr - tkptr->FindScript;
      if (Debug)
         fprintf (stdout, "%d |%s|\n",
                  tkptr->FindScriptLength, tkptr->FindScript);

      /* look backwards to see if a file type has been supplied */
      cptr = tkptr->FindScript;
      while (*sptr != '.' && *sptr != ']' && sptr > cptr) sptr--;
      if (*sptr != '.')
      {
         /* indicate no file type using 2 */
         tkptr->FindScriptCount = 2;
      }
      else
      {
         /**********************/
         /* file type supplied */
         /**********************/

         /* indicate file type supplied using 1 */
         tkptr->FindScriptCount = 1;
         tkptr->FindScriptTypePtr = sptr;
         tkptr->FindScriptTotalLength = tkptr->FindScriptLength;

         if (strsame (sptr, ".COM", -1))
            tkptr->RunTimePtr = "@";
         else
         if (strsame (sptr, ".EXE", -1))
            tkptr->RunTimePtr = "$";
         else
         {
            /* let's look for it in the run-time environment information */
            for (idx = 0; idx < Config.ScriptRunTimeCount; idx++)
            {
               if (Debug)
                  fprintf (stdout, "|%s|\n",
                           Config.ScriptRunTimeArray[idx]);
               sptr = tkptr->FindScriptTypePtr;
               cptr = Config.ScriptRunTimeArray[idx];
               while (*cptr && *cptr != ';' && *sptr)
               {
                  if (toupper(*cptr) != toupper(*sptr)) break;
                  cptr++;
                  sptr++;
               }
               if (*cptr == ';' && !*sptr) break;
            }

            if (idx >= Config.ScriptRunTimeCount)
            {
               /* nope, not found */
               static $DESCRIPTOR (ErrorScriptRunTimeFaoDsc,
"Execution of &nbsp;<TT>!AZ</TT>&nbsp; script types not configured.\0");
               char  String [256];
               $DESCRIPTOR (StringDsc, String);

               sys$fao (&ErrorScriptRunTimeFaoDsc, 0, &StringDsc, 
                        tkptr->FindScriptTypePtr);
               rqptr->ResponseStatusCode = 500;
               ErrorGeneral (rqptr, String, FI_LI);
               return (STS$K_ERROR);
            }

            /* found the file type, point to the run-time stuff */
            if (*cptr) cptr++;
            while (ISLWS(*cptr)) cptr++;
            tkptr->RunTimePtr = cptr;
            if (Debug) fprintf (stdout, "|%s|\n", tkptr->RunTimePtr);
         }
      }
   }

   if (tkptr->FindScriptCount == 2)
   {
      /*************************/
      /* first, DCL procedures */
      /*************************/

      memcpy (sptr = tkptr->FindScript+tkptr->FindScriptLength, ".COM", 5);
      tkptr->FindScriptTypePtr = sptr;
      tkptr->FindScriptTotalLength = tkptr->FindScriptLength + 4;
      tkptr->RunTimePtr = "@";
   }
   else
   if (tkptr->FindScriptCount == 3)
   {
      /***********************/
      /* second, executables */
      /***********************/

      memcpy (sptr = tkptr->FindScript+tkptr->FindScriptLength, ".EXE", 5);
      tkptr->FindScriptTypePtr = sptr;
      tkptr->FindScriptTotalLength = tkptr->FindScriptLength + 4;
      tkptr->RunTimePtr = "$";
   }
   else
   if (tkptr->FindScriptCount != 1 &&
       tkptr->FindScriptCount-4 < Config.ScriptRunTimeCount)
   {
      /*************************************/
      /* subsequent, run-time environments */
      /*************************************/

      tkptr->FindScriptTypePtr = sptr =
         tkptr->FindScript + tkptr->FindScriptLength;
      if (Debug)
         fprintf (stdout, "|%s|\n",
                  Config.ScriptRunTimeArray[tkptr->FindScriptCount-4]);
      for (cptr = Config.ScriptRunTimeArray[tkptr->FindScriptCount-4];
           *cptr && *cptr != ';';
           *sptr++ = *cptr++);
      *sptr = '\0';
      tkptr->FindScriptTotalLength = sptr - tkptr->FindScript;
      if (*cptr) cptr++;
      while (*cptr && ISLWS(*cptr)) cptr++;
      tkptr->RunTimePtr = cptr;
   }
   else
   if (tkptr->FindScriptCount != 1)
   {
      /**************/
      /* not found! */
      /**************/

      rqptr->ResponseStatusCode = 404;
      ErrorGeneral (rqptr, MsgFor(rqptr,MSG_SCRIPT_NOT_FOUND), FI_LI);
      return (STS$K_ERROR);
   }
   /* else 'tkptr->FindScriptCount' can be -1, when the type is supplied! */

   /********************/
   /* write search DCL */
   /********************/

   DclCommandDsc.dsc$a_pointer = tkptr->DclCommand;
   DclCommandDsc.dsc$w_length = sizeof(tkptr->DclCommand)-1;
   sys$fao (&SearchFaoDsc, &Length, &DclCommandDsc, tkptr->FindScript);
   if (Debug)
   {
      tkptr->DclCommand[Length] = '\0';      
      fprintf (stdout, "|%s|\n", tkptr->DclCommand);
   }

   DECnetWrite (rqptr, tkptr->DclCommand, Length);
   DECnetRead (rqptr);   
   return (SS$_RETRY);
}

/*****************************************************************************/
/*
Implement (most) OSU server scripting dialogs.
Behaviour determined from the OSU 'script_execute.c' module.
*/ 
 
DECnetOsuDialog (struct RequestStruct *rqptr)

{
   register int  cnt;
   register char  *cptr, *sptr, *zptr;
   register struct DECnetTaskStruct  *tkptr;

   int  value;
   char  Buffer [1024];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DECnetOsuDialog()\n");

   tkptr = rqptr->DECnetTaskPtr;

   if (tkptr->OsuDialogState == OSU_BEGIN)
   {
      /******************/
      /* initial dialog */
      /******************/

      /*
         OSU task executive expects four records as initial and unsolicited
         data from the server.  These are: script type, HTTP method name,
         HTTP protocol version, request path (excluding any query string).
      */

      tkptr->OsuDnetRecMode = tkptr->OsuDnetCgi = false;

      DECnetWrite (rqptr, "HTBIN", 5);
      DECnetWrite (rqptr, rqptr->HttpMethodName, -1);
      DECnetWrite (rqptr, HttpProtocol, -1);

      /* 
         URL (everything before any query string in original request)
         OSU requires the mapped script portion, which we have have to
         construct from the script name and the path information.
         "<DNETRQURL>" provides the unmapped (original) request path.
      */
      zptr = (sptr = Buffer) + sizeof(Buffer);
      cptr = rqptr->ScriptName;
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      cptr = rqptr->PathInfoPtr;
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      if (sptr >= zptr)
      {
         ErrorGeneralOverflow (rqptr, FI_LI);
         DECnetEnd (rqptr);
         return;
      }
      *sptr = '\0';
      /* it's an asynchronous write so we need some persistant storage */
      cptr = VmGetHeap (rqptr, sptr-Buffer+1);
      memcpy (cptr, Buffer, sptr-Buffer+1);
      DECnetWrite (rqptr, cptr, sptr-Buffer);

      /* now into dialog state */
      tkptr->OsuDialogState = OSU_DIALOG;

      DECnetRead (rqptr);
      return;
   }

   /**********************/
   /* dialog established */
   /**********************/

   cptr = rqptr->OutputBufferPtr;
   cnt = tkptr->DECnetReadIOsb.Count;
   cptr[cnt] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", cptr);

   if (tkptr->OsuDialogState == OSU_DIALOG && *cptr != '<')
   {
      /************************/
      /* start of data stream */
      /************************/

      if (tkptr->OsuDnetCgi)
      {
         /* CGI script, check the initial response */
         value = CgiScriptHeader (rqptr, cptr, cnt);
         if (value == CGI_OUTPUT_TERMINATE)
         {
            /* terminate processing */
            DECnetEnd (rqptr);
            return;
         }
      }

      /* now set the carriage-control appropriately */
      if (tkptr->OsuDnetRecMode)
         tkptr->OsuDialogState = OSU_OUTPUT_REC;
      else
         tkptr->OsuDialogState = OSU_OUTPUT_RAW;

      /* drop through so the line can be processed by "_REC" or "_RAW" */
   }

   switch (tkptr->OsuDialogState)
   {
      case OSU_DIALOG :

         /********************/
         /* stream massaging */
         /********************/

         if (!memcmp (cptr, "<DNETRECMODE>", 13))
         {
            /* forces text mode */
            tkptr->OsuDnetRecMode = true;
            DECnetRead (rqptr);
            return;
         }

         if (!memcmp (cptr, "<DNETCGI>", 9))
         {
            /* similar to raw mode except response header is checked */
            tkptr->OsuDnetCgi = true;
            DECnetRead (rqptr);
            return;
         }

         if (!memcmp (cptr, "<DNETRAW>", 9))
         {
            /* script handles all response formatting, return byte-for-byte */
            DECnetRead (rqptr);
            return;
         }

         if (!memcmp (cptr, "<DNETTEXT>", 10))
         {
            /* script sends status line then plain text */
            tkptr->OsuDialogState = OSU_DNETTEXT;
            DECnetRead (rqptr);
            return;
         }

         if (!memcmp (cptr, "</DNETCGI>", 10) ||
             !memcmp (cptr, "</DNETRAW>", 10) ||
             !memcmp (cptr, "</DNETTEXT>", 11))
         {
            /* end of script output */
            DECnetEnd (rqptr);
            return;
         }

         /**********************/
         /* information dialog */
         /**********************/

         if (!memcmp (cptr, "<DNETARG>", 9))
         {
            /* full search argument (query string) */
            cptr = VmGetHeap (rqptr, rqptr->QueryStringLength+2);
            /* OSU supplies a leading '?' query separator */
            *cptr = '?';
            memcpy (cptr+1, rqptr->QueryStringPtr, rqptr->QueryStringLength);
            DECnetWrite (rqptr, cptr, rqptr->QueryStringLength+1);
            DECnetRead (rqptr);
            return;
         }

         if (!memcmp (cptr, "<DNETARG2>", 10))
         {
            /* truncated search argument (query string) */
            cptr = VmGetHeap (rqptr, 256);
            /* OSU supplies a leading '?' query separator */
            *cptr = '?';
            if (rqptr->QueryStringLength > 254)
            {
               memcpy (cptr+1, rqptr->QueryStringPtr, 254);
               cptr[255] = '\0';
               DECnetWrite (rqptr, cptr, 255);
            }
            else
            {
               memcpy (cptr+1, rqptr->QueryStringPtr, rqptr->QueryStringLength);
               DECnetWrite (rqptr, cptr, rqptr->QueryStringLength+1);
            }
            DECnetRead (rqptr);
            return;
         }

         if (!memcmp (cptr, "<DNETPATH>", 10))
         {
            /* script path (but excluding script name itself) */
            register char  *cptr;
            for (cptr = rqptr->ScriptName; *cptr; cptr++);
            while (cptr > rqptr->ScriptName && *cptr != '/') cptr--;
            DECnetWrite (rqptr, rqptr->ScriptName,
                         cptr-(char*)rqptr->ScriptName+1);
            DECnetRead (rqptr);
            return;
         }

         if (!memcmp (cptr, "<DNETHDR>", 9))
         {
            /* send the full request header */
            tkptr->OsuDialogState = OSU_DNET_HDR;
            DECnetWriteRequest (rqptr);
            return;
         }

         if (!memcmp (cptr, "<DNETINPUT>", 11))
         {
            /* begin to write the request body */
            tkptr->OsuDialogState = OSU_DNET_INPUT;
            DECnetWriteRequest (rqptr);
            return;
         }

         if (!memcmp (cptr, "<DNETBINDIR>", 12))
         {
            /*
               OSU script and binaries VMS directory specification.
               Isolate the 'HT_ROOT:[dir]' from the example
               'node::"0=wwwexec"HT_ROOT:[dir]script.ext'
            */
            for (cptr = tkptr->MappedScript; *cptr && *cptr != ':'; cptr++);
            if (*cptr == ':') cptr++;
            if (*cptr == ':') cptr++;
            if (*cptr == '\"')
            {
               cptr++;
               while (*cptr && *cptr != '\"') cptr++;
               if (*cptr) cptr++;
            }
            sptr = cptr;
            while (*cptr) cptr++;
            while (cptr > tkptr->MappedScript && *cptr != ']') cptr--;
            DECnetWrite (rqptr, sptr, cptr - sptr + 1);
            DECnetRead (rqptr);
            return;
         }

         if (!memcmp (cptr, "<DNETRQURL>", 11))
         {
            /* full request URL prior to any mapping */
            DECnetWrite (rqptr, rqptr->ResourcePtr, -1);
            DECnetRead (rqptr);
            return;
         }

         if (!memcmp (cptr, "<DNETINVCACHE>", 14))
         {
            /* invalidate cache */
            rqptr->ResponseStatusCode = 501;
            ErrorGeneral (rqptr, ErrorOsuNoInvCache, FI_LI);
            DECnetEnd (rqptr);
            return;
         }

         if (!memcmp (cptr, "<DNETHOST>", 10))
         {
            /* server host name */
            DECnetWrite (rqptr, rqptr->ServicePtr->ServerHostName, -1);
            DECnetRead (rqptr);
            return;
         }

         if (!memcmp (cptr, "<DNETID>", 8))
         {
            /* composite line of request data */
            DECnetOsuDnetId (rqptr, tkptr, false);
            return;
         }

         if (!memcmp (cptr, "<DNETID2>", 9))
         {
            /* composite line of request data (plus a bit!) */
            DECnetOsuDnetId (rqptr, tkptr, true);
            return;
         }

         if (!memcmp (cptr, "<DNETMANAGE>", 12))
         {
            /* server management */
            rqptr->ResponseStatusCode = 500;
            ErrorGeneral (rqptr, ErrorOsuNoManage, FI_LI);
            DECnetEnd (rqptr);
            return;
         }

         if (!memcmp (cptr, "<DNETXLATE>", 11) ||
             !memcmp (cptr, "<DNETXLATEV>", 12))
         {
            /* into translate state */
            tkptr->OsuDialogState = OSU_DNET_XLATE;
            /* read the path to be translated */
            DECnetRead (rqptr);
            return;
         }


      /****************/
      /* other states */
      /****************/

      case OSU_DNETTEXT :

         /*
            This output record contains a status line something like "200 ok".
            Get the status code, ignore the rest, generate an HTTP header.
            Then move into a state to accept record-oriented script output.
         */
         while (*cptr && !isdigit(*cptr) && *cptr != '\n') cptr++;
         if (isdigit(*cptr)) rqptr->ResponseStatusCode = atoi(cptr);
         if ((rqptr->ResponseHeaderPtr =
              HttpHeader (rqptr, rqptr->ResponseStatusCode,
                          "text/plain", -1, NULL, NULL)) == NULL)
         {
            DECnetEnd (rqptr);
            return;
         }

         /* now move into record output state */
         tkptr->OsuDialogState = OSU_OUTPUT_REC;
         DECnetRead (rqptr);

         return;


      case OSU_OUTPUT_RAW :

         /*
            This script output record is "binary" data.  Write it to
            the client via the network without any further processing.
         */
         if (!memcmp (cptr, "</DNETCGI>", 10) ||
             !memcmp (cptr, "</DNETRAW>", 10) ||
             !memcmp (cptr, "</DNETTEXT>", 11))
         {
            /* end of script output */
            DECnetEnd (rqptr);
            return;
         }

         NetWrite (rqptr, &DECnetRead, cptr, cnt);
         return;


      case OSU_OUTPUT_REC :

         /*
            This record of script output is essentially a line of "text".
            Esure it has correct HTTP carriage-control (trailing newline).
            Write this (possibly) massaged record to the client (network).
         */
         if (!memcmp (cptr, "</DNETCGI>", 10) ||
             !memcmp (cptr, "</DNETRAW>", 10) ||
             !memcmp (cptr, "</DNETTEXT>", 11))
         {
            /* end of script output */
            DECnetEnd (rqptr);
            return;
         }

         /* adjust carriage-control if necessary */
         if (cnt)
            if (cptr[cnt-1] != '\n')
               cptr[cnt++] = '\n';
            else;
         else
            cptr[cnt++] = '\n';
         if (Debug) cptr[cnt] = '\0';
         NetWrite (rqptr, &DECnetRead, cptr, cnt);
         return;


      case OSU_DNET_XLATE :

         /* record contains a URL-style path to be translated into RMS */
         DECnetOsuDnetXlate (rqptr, tkptr, cptr);
         return;


      default :

         /* Hmmm, problem? */
         rqptr->ResponseStatusCode = 500;
         ErrorGeneral (rqptr, ErrorOsuImplementation, FI_LI);
         DECnetEnd (rqptr);
         return;
   }
}

/*****************************************************************************/
/*
Write a single record comprising a series of white-space separated request
data to the OSU task.
*/ 
 
DECnetOsuDnetId
(
struct RequestStruct *rqptr,
struct DECnetTaskStruct *tkptr,
boolean Id2
)
{
   static $DESCRIPTOR (SignedLongFaoDsc, "!SL\0");
   register char  *cptr, *sptr, *zptr;

   char Buffer [1024],
        String [256];
   $DESCRIPTOR (StringDsc, String);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DECnetOsuDnetId()\n");

   cptr = rqptr->OutputBufferPtr;
   zptr = (sptr = Buffer) + sizeof(Buffer);

   /* server identification (underscores substituted for spaces) */
   cptr = SoftwareID;
   while (*cptr && sptr < zptr)
   {
      if (*cptr == ' ')
      {
         *sptr++ = '_';
         cptr++;
      }
      else
         *sptr++ = *cptr++;
   }

   /* local host */
   if (sptr < zptr) *sptr++ = ' ';
   for (cptr = rqptr->ServicePtr->ServerHostName;
        *cptr && sptr < zptr; *sptr++ = *cptr++);

   /* local port */
   if (sptr < zptr) *sptr++ = ' ';
   for (cptr = rqptr->ServicePtr->ServerPortString;
        *cptr && sptr < zptr; *sptr++ = *cptr++);

   /* remote port (not available for WASD) */
   if (sptr < zptr) *sptr++ = ' ';
   for (cptr = "65535"; *cptr && sptr < zptr; *sptr++ = *cptr++);

   /* remote host address */
   if (sptr < zptr) *sptr++ = ' ';
   sys$fao (&SignedLongFaoDsc, 0, &StringDsc,
            rqptr->ClientInternetAddress32bit);
   for (cptr = String; *cptr && sptr < zptr; *sptr++ = *cptr++);

   /* remote user */
   if (rqptr->RemoteUser[0])
   {
      if (sptr < zptr) *sptr++ = ' ';
      for (cptr = rqptr->RemoteUser;
           *cptr && sptr < zptr; *sptr++ = *cptr++);
   }
   else
   if (Id2)
      if (sptr < zptr) *sptr++ = ' ';

   if (Id2 && Config.DnsLookup)
   {
      /* remote host name */
      if (sptr < zptr) *sptr++ = ' ';
      for (cptr = rqptr->ClientHostName;
           *cptr && sptr < zptr; *sptr++ = *cptr++);
   }

   *sptr = '\0';

   /* it's an asynchronous write so we need some persistant storage */
   cptr = VmGetHeap (rqptr, sptr-Buffer+1);
   memcpy (cptr, Buffer, sptr-Buffer+1);

   DECnetWrite (rqptr, cptr, sptr-Buffer);
   DECnetRead (rqptr);
}

/*****************************************************************************/
/*
Map the supplied path (either absolute or relative to the request path) into a
VMS file system specification and write that to the OSU task.
*/ 
 
DECnetOsuDnetXlate
(
struct RequestStruct *rqptr,
struct DECnetTaskStruct *tkptr,
char *Path
)
{
   register char  *cptr;

   char AbsPath [256],
        Vms [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DECnetOsuDnetXlate() |%s|\n", Path);

   /* WASD always ignores any method */
   if (*(cptr = Path) == '(')
   {
      while (*cptr && *cptr != ')') cptr++;
      if (*cptr) cptr++;
   }

   Vms[0] = '\0';
   if (*cptr == '.')
   {
      MapUrl_VirtualPath (rqptr->PathInfoPtr, cptr, AbsPath, sizeof(AbsPath));
      cptr = AbsPath;
   }
   if (*cptr) MapUrl_Map (cptr, Vms, NULL, NULL, rqptr);

   DECnetWrite (rqptr, Vms, -1);
   DECnetRead (rqptr);

   /* back from translate into dialog state */
   tkptr->OsuDialogState = OSU_DIALOG;
}

/*****************************************************************************/

